#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <curses.h>
#include <err.h>
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "dvview.h"
#include "ctrl/usb_ctrl.h"

static const char rcsid[] = "$Id: dvview.c,v 1.8 2004/07/29 01:31:25 a-gota Exp $";

#define	max(a, b)	(a < b ? b : a)

#define	PLAYDV_PATH	"/usr/local/bin/playdv"
#define	DIF_BLK_SIZE		    80
#define	DIF_NUM_BLKS_IN_SEQ	   150
#define	DIF_SEQ_SIZE		(DIF_BLK_SIZE * DIF_NUM_BLKS_IN_SEQ)
#define	NUM_DIF_SEQ_525		    10
#define	NUM_DIF_SEQ_625		    12
#define	MAX_FRAME_SIZE	(DIF_SEQ_SIZE * max(NUM_DIF_SEQ_525, NUM_DIF_SEQ_625))

#define	UHID_DEVNAME	"/dev/uhid0"
#define	UHID_DEVTYPE	"CommandPost"

#define	NUM_PGM		10

struct cut_pos {
    long in, out;
};

#define	NUM_BANK	 8

struct stream_bank {
    int			no;
    char		*filename;
    FILE		*fp;
    int			system_is_625;
    long		frame_size, num_frame;
    struct cut_pos	cut_pos[NUM_PGM];
    int			last_pos;
    struct timeval	*timeout;
} bank[NUM_BANK], *cur_bank = NULL;

struct timeval	timeout_0, timeout_50, timeout_60;
u_char *frame = NULL;
pid_t pid = 0;
WINDOW *win_action = NULL, *win_frame_info = NULL, *win_status = NULL, *win_streams = NULL, *win_dialog = NULL;
FILE *fp_stdin;
struct termios t_orig, t_raw;
char *recdate, *rectime, *source;


void dumb_print(char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
    putchar('\n');
}


void dumb_print_nonl(char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
}


void dumb_print_term(char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
    putchar('\n');
}


void dumb_put_frame_no(long frame_no) {
    printf("#%05ld", frame_no);
}


void dumb_put_timecode(int hh, int mm, int ss, int fr, int sys_625, int df) {
    printf(": %02d:%02d:%02d.%02d", hh, mm, ss, fr);
    if (!sys_625) {
	fputs(df ? " DF" : " NDF", stdout);
    }
}


void dumb_put_datacode(char *recdate, char *rectime, char *source) {
    fputs("    ", stdout);
    fputs(recdate, stdout);
    putchar(' ');
    fputs(rectime, stdout);
    fputs("  ", stdout);
    fputs(source, stdout);
}


void dumb_finish_frame_info(void) {
    putchar('\r');
    fflush(stdout);
}


void dumb_clear_frame_info(void) {
    putchar('\n');
}


void dumb_warn(char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vwarn(fmt, ap);
    va_end(ap);
}


void dumb_clear_status(void) {
    putchar('\n');
}


void dumb_put_stream_info(struct stream_bank *bank) {
    if (bank->fp != NULL) {
	printf("DV stream #%d is opened; %s, ", bank->no + 1, bank->filename);
	if (bank->system_is_625 == 0) {
	    printf("525/60 system");
	} else {
	    printf("625/50 system");
	}
	printf(", %ld frame(s)\n", bank->num_frame);
    }
}


void dumb_put_selected_stream(struct stream_bank *bank, int selected) {
    return;
}


void dumb_get_filename(int bankno, char *buf, int len) {
    tcsetattr(fileno(fp_stdin), TCSAFLUSH, &t_orig);
    printf("\nOpen file\nDV stream file #%d: ", bankno + 1);
    fflush(stdout);
    fgets(buf, len, fp_stdin);
    tcsetattr(fileno(fp_stdin), TCSAFLUSH, &t_raw);
}


void dumb_clean_up(void) {
    printf("\nClean up...\n");
}


void curses_put_action_norefresh(char *fmt, ...) {
    va_list ap;

    wmove(win_action, 0, 0);
    va_start(ap, fmt);
    vw_printw(win_action, fmt, ap);
    va_end(ap);
}


void curses_put_action(char *fmt, ...) {
    va_list ap;

    wmove(win_action, 0, 0);
    va_start(ap, fmt);
    vw_printw(win_action, fmt, ap);
    va_end(ap);
    wclrtoeol(win_action);
    wrefresh(win_action);
}


void curses_add_action_norefresh(char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vw_printw(win_action, fmt, ap);
    va_end(ap);
}


void curses_add_action(char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vw_printw(win_action, fmt, ap);
    va_end(ap);
    wclrtoeol(win_action);
    wrefresh(win_action);
}


void curses_put_frame_no(long frame_no) {
    mvwprintw(win_frame_info, 0, 0, "#%05ld", frame_no);
    wclrtoeol(win_frame_info);
}


void curses_put_timecode(int hh, int mm, int ss, int fr, int sys_625, int df) {
    mvwprintw(win_frame_info, 1, 0, "%02d:%02d:%02d.%02d", hh, mm, ss, fr);
    if (!sys_625) {
	waddstr(win_frame_info, df ? " DF" : " NDF");
    }
    wclrtoeol(win_frame_info);
}


void curses_put_datacode(char *recdate, char *rectime, char *source) {
    mvwaddstr(win_frame_info, 3, 2, recdate);
    mvwaddstr(win_frame_info, 4, 2, rectime);
    mvwaddstr(win_frame_info, 5, 2, source);
    wclrtoeol(win_frame_info);
}


void curses_finish_frame_info(void) {
    wrefresh(win_frame_info);
}


void curses_clear_frame_info(void) {
    wclear(win_frame_info);
    wrefresh(win_frame_info);
}


void curses_put_status_norefresh(char *fmt, ...) {
    va_list ap;

    wmove(win_status, 0, 0);
    va_start(ap, fmt);
    vw_printw(win_status, fmt, ap);
    va_end(ap);
}


void curses_put_status(char *fmt, ...) {
    va_list ap;

    wmove(win_status, 0, 0);
    va_start(ap, fmt);
    vw_printw(win_status, fmt, ap);
    va_end(ap);
    wclrtoeol(win_status);
    wrefresh(win_status);
}


void curses_add_status_norefresh(char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vw_printw(win_status, fmt, ap);
    va_end(ap);
}


void curses_add_status(char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vw_printw(win_status, fmt, ap);
    va_end(ap);
    wclrtoeol(win_status);
    wrefresh(win_status);
}


void curses_put_warn(char *fmt, ...) {
    va_list ap;

    wmove(win_status, 0, 0);
    va_start(ap, fmt);
    vw_printw(win_status, fmt, ap);
    va_end(ap);
    waddstr(win_status, ": ");
    waddstr(win_status, strerror(errno));
    wclrtoeol(win_status);
    wrefresh(win_status);
}


void curses_clear_status(void) {
    wclear(win_status);
    wrefresh(win_status);
}


void curses_show_stream_info(struct stream_bank *bank) {
    if (bank->fp != NULL) {
	mvwprintw(win_status, 0, 0, "DV stream #%d is opened", bank->no + 1);
	wclrtoeol(win_status);
	wrefresh(win_status);
    }
    mvwprintw(win_streams, bank->no + 3, 3, "%2d: ", bank->no + 1);
    if (bank->fp != NULL) {
	waddstr(win_streams, bank->filename);
	waddstr(win_streams, ", ");
	if (bank->system_is_625 == 0) {
	    waddstr(win_streams, "525/60");
	} else {
	    waddstr(win_streams, "625/50");
	}
	wprintw(win_streams, " system, %ld frame(s)", bank->num_frame);
    }
    wclrtoeol(win_streams);
    wrefresh(win_streams);
}


void curses_mark_stream_selected(struct stream_bank *bank, int selected) {
    mvwaddch(win_streams, bank->no + 3, 1, selected ? '>' : ' ');
    wrefresh(win_streams);
}


void curses_filename_dialog(int bankno, char *buf, int len) {
    WINDOW *dialog;
    int ret;

    if ((win_dialog = newwin(4, COLS - 10, (LINES - 4) / 2, 10 / 2)) == NULL) {
	err(3, "newwin(dialog)");
    }
    box(win_dialog, 0, 0);
    mvwaddstr(win_dialog, 0, 2, "Open file");
    mvwprintw(win_dialog, 1, 2, "DV stream file #%d:", bankno + 1);
    wrefresh(win_dialog);
    if ((dialog = derwin(win_dialog, 2, COLS - 10 - 22 - 2, 1, 22)) == NULL) {
	err(4, "derwin(dialog)");
    }
    scrollok(dialog, TRUE);
    touchwin(win_dialog);
    wrefresh(dialog);

    nocbreak();
    echo();
    ret = mvwgetnstr(dialog, 0, 0, buf, len);
    noecho();
    cbreak();

    delwin(dialog);
    delwin(win_dialog);
    win_dialog = NULL;

    touchwin(stdscr);
    refresh();
}


void curses_clean_up(void) {
    wclear(win_action);
    wrefresh(win_action);
    wclear(win_frame_info);
    wrefresh(win_frame_info);
    wclear(win_status);
    wrefresh(win_status);
    wclear(win_streams);
    wrefresh(win_streams);
    endwin();
    printf("Clean up...\n");
}

void do_nothing(void) {
    return;
}

struct cons_handler {
    void (*put_action)(char *fmt, ...);
    void (*put_action_nonl)(char *fmt, ...);
    void (*put_action_cont)(char *fmt, ...);
    void (*put_action_term)(char *fmt, ...);
    void (*put_frame_no)(long frame_no);
    void (*disp_timecode)(int hh, int mm, int ss, int fr, int sys_625, int df);
    void (*disp_datacode)(char *recdate, char *rectime, char *source);
    void (*finish_frame_info)(void);
    void (*clear_frame_info)(void);
    void (*put_status)(char *fmt, ...);
    void (*put_status_nonl)(char *fmt, ...);
    void (*put_status_cont)(char *fmt, ...);
    void (*put_status_term)(char *fmt, ...);
    void (*warn)(char *fmt, ...);
    void (*clear_status)(void);
    void (*show_stream_info)(struct stream_bank *bank);
    void (*mark_stream_selected)(struct stream_bank *bank, int selected);
    void (*filename_dialog)(int bankno, char *buf, int len);
    void (*clean_up)(void);
} *cons, dumb_handler = {
    dumb_print,
    dumb_print_nonl,
    dumb_print_nonl,
    dumb_print_term,
    dumb_put_frame_no,
    dumb_put_timecode,
    dumb_put_datacode,
    dumb_finish_frame_info,
    dumb_clear_frame_info,
    dumb_print,
    dumb_print_nonl,
    dumb_print_nonl,
    dumb_print_term,
    dumb_warn,
    dumb_clear_status,
    dumb_put_stream_info,
    dumb_put_selected_stream,
    dumb_get_filename,
    dumb_clean_up
}, curses_handler = {
    curses_put_action,
    curses_put_action_norefresh,
    curses_add_action_norefresh,
    curses_add_action,
    curses_put_frame_no,
    curses_put_timecode,
    curses_put_datacode,
    curses_finish_frame_info,
    curses_clear_frame_info,
    curses_put_status,
    curses_put_status_norefresh,
    curses_add_status_norefresh,
    curses_add_status,
    curses_put_warn,
    curses_clear_status,
    curses_show_stream_info,
    curses_mark_stream_selected,
    curses_filename_dialog,
    curses_clean_up
};


void
clean_up(void)
{
    int tmp;

    cons->clean_up();
    for (tmp = 0; tmp < NUM_BANK; tmp++) {
	if (bank[tmp].filename != NULL) {
	    free(bank[tmp].filename);
	}
	if (bank[tmp].fp != NULL) {
	    fclose(bank[tmp].fp);
	}
    }
    if (frame != NULL) {
	free(frame);
    }
    if (0 < pid) {
	kill(pid, SIGTERM);
	waitpid(pid, &tmp, WNOHANG);
    }
}


int
main(int argc, char *argv[])
{
    char filenamebuf[FILENAME_MAX], *ptr;
    int num_opened, bankno;
    FILE *fp;
    int controller_fd, num_fd;
    int command, play_f, eof_f, exit_f = 0, pipefd[2], verbose_f = 0;
    int use_curses_f = 0;
    int param, skip, div, wait_frame;
    long frame_size, num_frame, frame_no = 0;
    fd_set readfds;
    int c, nodisp_f = 0, error_f = 0;
    char *playdv_path = PLAYDV_PATH;
    char *uhid_path = UHID_DEVNAME, *uhid_type = UHID_DEVTYPE;

    FILE *open_stream(struct stream_bank *bank, const char *filename);
    int get_key_event(FILE *fp);
    void parse_subcode(u_char *frame, int system_is_625);
    void usage(char *argv0);

    while ((c = getopt(argc, argv, "c:hnp:vw")) != EOF) {
	switch (c) {
	  case 'c':
	    if ((ptr = strchr(optarg, ':')) != NULL) {
		*ptr = '\0';
		uhid_type = ptr + 1;
	    }
	    if (optarg != ptr) {
		uhid_path = optarg;
	    }
	    break;
	  case 'h':
	    ++error_f;
	    break;
	  case 'n':
	    ++nodisp_f;
	    break;
	  case 'p':
	    playdv_path = optarg;
	    break;
	  case 'v':
	    ++verbose_f;
	    break;
	  case 'w':
	    ++use_curses_f;
	    break;
	  default:
	    ++error_f;
	    break;
	}
    }

    if (error_f) {
	usage(argv[0]);
    }

    argc -= optind;
    argv += optind;

    pipe(pipefd);
    if ((pid = fork()) == 0) {
	/* child */
	dup2(pipefd[0], 0);
	close(pipefd[0]);
	execl(playdv_path, playdv_path, NULL);
	return 1;
    }

    if ((frame = malloc(MAX_FRAME_SIZE)) == NULL) {
	kill(pid, SIGTERM);
	waitpid(pid, &command, WNOHANG);
	err(1, "malloc for frame memory");
    }

    controller_fd = controller_init(uhid_path);

    fp_stdin = fopen("/dev/tty", "r");
    if (use_curses_f) {
	initscr();
	newterm(getenv("TERM"), stdout, fp_stdin);
	savetty();
	noecho();
	cbreak();
	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
	if ((win_action = subwin(stdscr, 1, 20, 0, 0)) == NULL) {
	    err(2, "newwin(action)");
	}
	touchwin(stdscr);
	wrefresh(win_action);
	if ((win_frame_info = subwin(stdscr, 6, 20, 2, 0)) == NULL) {
	    err(2, "newwin(frame_info)");
	}
	touchwin(stdscr);
	wrefresh(win_frame_info);
	if ((win_streams = subwin(stdscr,
				  NUM_BANK + 4, COLS,
				  LINES - (NUM_BANK + 4 + 1), 0)) == NULL) {
	    err(2, "newwin(streams)");
	}
	wborder(win_streams, ' ', ' ', '-', '-', '-', '-', '-', '-');
	mvwaddstr(win_streams, 1, 0, "Opened stream (`>': selected)");
	touchwin(stdscr);
	wrefresh(win_streams);
	if ((win_status = subwin(stdscr, 1, COLS, LINES - 1, 0)) == NULL) {
	    err(2, "newwin(status)");
	}
	touchwin(stdscr);
	wrefresh(win_status);
	cons = &curses_handler;
    } else {
	tcgetattr(fileno(fp_stdin), &t_orig);
	t_raw = t_orig;
	t_raw.c_lflag &= ~(ECHO|ICANON);
	tcsetattr(fileno(fp_stdin), TCSAFLUSH, &t_raw);
	cons = &dumb_handler;
    }

    bzero(bank, sizeof(struct stream_bank) * NUM_BANK);
    num_opened = 0;
    for (cur_bank = bank; 1 <= argc; argc--) {
	cur_bank->no = num_opened;
	if (open_stream(cur_bank, *argv++) != NULL) {
	    ++cur_bank;
	    if (NUM_BANK <= ++num_opened) {
		break;
	    }
	}
    }
    bankno = 0;
    cur_bank = bank;
    fp = cur_bank->fp;
    frame_size = cur_bank->frame_size;
    num_frame = cur_bank->num_frame;
    cons->mark_stream_selected(cur_bank, 1);

    timeout_0.tv_sec = 0;
    timeout_0.tv_usec = 0;
    timeout_50.tv_sec = 0;
    timeout_50.tv_usec = 20000;
    timeout_60.tv_sec = 0;
    timeout_60.tv_usec = 16666;

    if (num_opened == 0) {
	cons->put_status("No DV stream file is opened");
    } else {
	cons->put_status("%d streams are opened", num_opened);
    }

    play_f = 0;
    skip = 0;
    div = 1;
    wait_frame = 0;
    eof_f = 0;
    exit_f = 0;

    num_fd = max(controller_fd, fileno(fp_stdin)) + 1;

    cons->put_action("Stop");

    do {
	FD_ZERO(&readfds);
	FD_SET(fileno(fp_stdin), &readfds);
	if (0 < controller_fd) {
	    FD_SET(controller_fd, &readfds);
	}
	select(num_fd, &readfds, NULL, NULL,
	       play_f ? (1 < div ? bank[bankno].timeout : &timeout_0) : NULL);
	command = ACT_NONE;
	if (FD_ISSET(fileno(fp_stdin), &readfds)) {
	    command = get_key_event(fp_stdin);
	}
	if (command == ACT_NONE && 0 < controller_fd && FD_ISSET(controller_fd, &readfds)) {
	    command = controller_get_event(controller_fd);
	}
	if (command != ACT_NONE) {
	    cons->clear_status();
	    param = (char)(command & ACT_PARAM);

	    if (fp != NULL) {
		switch (command & ACT_MASK) {
		  case ACT_PAUSE:
		    if (play_f) {
			play_f = -1;
		    }
		    div = 1;
		    skip = 0;
		    cons->put_action("Pause");
		    break;
		  case ACT_PLAY:
		    if (param == 0) {
			if (play_f) {
			    if (div == 1 && skip == 0) {
				play_f = 0;
				cons->put_action("Pause");
				break;
			    }
			} else {
			    if (eof_f) {
				fseek(fp, 0L, SEEK_SET);
				eof_f = 0;
				frame_no = 0L;
			    }
			}
			param = 1;
		    }
		    cons->put_action_nonl("Play ");
		    play_f = 1;
		    if (0 < param) {
			skip = param - 1;
			div = 1;
			cons->put_action_term("X%d", param);
		    } else {
			skip = 0;
			div = -param;
			cons->put_action_term("1/%d", -param);
		    }
		    break;
		  case ACT_BACK:
		    cons->put_action_nonl("Backward Play ");
		    play_f = 1;
		    if (0 < param) {
			skip = -(param + 1);
			div = 1;
			cons->put_action_term("X%d", param);
		    } else {
			skip = -2;
			div = -param;
			cons->put_action_term("1/%d", -param);
		    }
		    break;
		  case ACT_STOP:
		    cons->put_action("Stop");
		    play_f = -1;
		    skip = 0;
		    fseek(fp, 0L, SEEK_SET);
		    break;
		  case ACT_NEXT:
		    cons->put_action("Next frame");
		    play_f = -1;
		    skip = 0;
		    break;
		  case ACT_PREV:
		    cons->put_action("Previous frame");
		    play_f = -1;
		    skip = 0;
		    if (eof_f) {
			fseek(bank[bankno].fp, frame_size * (num_frame - 1), SEEK_SET);
			eof_f = 0;
		    } else {
			if ((frame_no -= 2) < 0) {
			    frame_no = 0L;
			    fseek(bank[bankno].fp, 0L, SEEK_SET);
			} else {
			    fseek(bank[bankno].fp, frame_size * -2, SEEK_CUR);
			}
		    }
		    break;
		  case ACT_REFRESH:
		    cons->put_action("Refresh last frame");
		    play_f = -1;
		    skip = 0;
		    if (eof_f) {
			frame_no = num_frame - 1;
			fseek(bank[bankno].fp, frame_size * frame_no, SEEK_SET);
			eof_f = 0;
		    } else {
			if (--frame_no < 0) {
			    frame_no = 0L;
			    fseek(bank[bankno].fp, 0L, SEEK_SET);
			} else {
			    fseek(bank[bankno].fp, frame_size * -1, SEEK_CUR);
			}
		    }
		    break;
		}
	    }
	    switch (command & ACT_MASK) {
	      case ACT_OPEN:
		play_f = 0;
		cons->put_action(frame_no == 0 ? "Stop" : "Pause");
		cons->clear_frame_info();
		if (0 < param && param <= NUM_BANK) {
		    cons->mark_stream_selected(&bank[bankno], 0);
		    if (eof_f) {
			bank[bankno].last_pos = 0L;
		    } else {
			bank[bankno].last_pos = frame_no - 1;
		    }
		    bankno = param - 1;
		    cur_bank = &bank[bankno];
		    cur_bank->no = bankno;
		    fp = cur_bank->fp;
		    frame_size = cur_bank->frame_size;
		    num_frame = cur_bank->num_frame;
		    if ((frame_no = cur_bank->last_pos) < 0) {
			frame_no = 0L;
		    }
		    cons->mark_stream_selected(cur_bank, 1);
		    if (fp != NULL) {
			fseek(fp, frame_size * frame_no, SEEK_SET);
			cons->put_status("DV stream is switched to #%d; "
				 "%s, %ld @ %ld frames\n",
			       bankno + 1, bank[bankno].filename,
			       frame_no, num_frame);
			eof_f = 0;
			cons->put_action(frame_no == 0 ? "Stop" : "Pause");
			break;
		    }
		} else {
		    if (bank[bankno].fp != NULL) {
			fclose(bank[bankno].fp);
			fp = bank[bankno].fp = NULL;
			if (bank[bankno].filename != NULL) {
			    free(bank[bankno].filename);
			    bank[bankno].filename = NULL;
			}
			--num_opened;
		    }
		}
		cons->show_stream_info(&bank[bankno]);
		cons->filename_dialog(bankno, filenamebuf, FILENAME_MAX - 1);
		if ((ptr = strchr(filenamebuf, '\n')) != NULL) {
		    *ptr = '\0';
		}
		bank[bankno].no = bankno;
		cur_bank = &bank[bankno];
		if (open_stream(cur_bank, filenamebuf) != NULL) {
		    ++num_opened;
		    fp = cur_bank->fp;
		    frame_size = cur_bank->frame_size;
		    num_frame = cur_bank->num_frame;
		    frame_no = 0L;
		}
		play_f = 0;
		cons->put_action(frame_no == 0 ? "Stop" : "Pause");
		skip = 0;
		break;
	      case ACT_DISP:
		if ((verbose_f = !verbose_f)) {
		    cons->put_status("Verbose: On");
		} else {
		    cons->put_status("Verbose: Off");
		    cons->clear_frame_info();
		}
		break;
	      case ACT_QUIT:
		play_f = 0;
		exit_f = 1;
		break;
	    }
	}

	if (play_f && fp == NULL) {
	    cons->put_status("No DV stream file is assumed");
	    play_f = 0;
	}

	if (play_f) {
	    if (wait_frame++ < div) {
		continue;
	    }
	    wait_frame = 0;
	    if (fp == NULL) {
		cons->put_status("Not opened");
		++eof_f;
		play_f = -1;
		cons->put_action("Stop");
	    } else if (fread(frame, frame_size, 1, fp) < 1) {
		cons->put_status("EOF      ");
		++eof_f;
		play_f = -1;
		cons->put_action("Stop");
	    } else {
		cons->put_frame_no(frame_no);
		if (verbose_f) {
		    parse_subcode(frame, bank[bankno].system_is_625);
		}
		cons->finish_frame_info();
		if (!nodisp_f) {
		    if (write(pipefd[1], frame, frame_size) < frame_size) {
			++exit_f;
		    }
		}
		if ((frame_no += (skip + 1)) < 0) {
		    frame_no = 0L;
		    fseek(fp, 0L, SEEK_SET);
		    play_f = 0;
		    cons->put_action("Pause");
		} else if (skip != 0) {
		    fseek(fp, frame_size * skip, SEEK_CUR);
		}
	    }
	    if (play_f < 0) {
		++play_f;
	    }
	}
    } while (!exit_f);

    return 0;
}

FILE
*open_stream(struct stream_bank *bank, const char *filename)
{
    int cnt;
    u_char frame[DIF_BLK_SIZE];
    struct stat stat;

    if (bank->filename != NULL) {
	free(bank->filename);
    }
    bank->filename = strdup(filename);
    if ((bank->fp = fopen(bank->filename, "r")) == NULL) {
	cons->warn("%s", bank->filename);
    } else {
	fread(frame, DIF_BLK_SIZE, 1, bank->fp);
	if ((frame[0] & 0xe0) != 0) {
	    /* Header is not found */
	    cons->put_status("%s: broken DV stream", bank->filename);
	    fclose(bank->fp);
	    bank->fp = NULL;
	} else {
	    fseek(bank->fp, 0L, SEEK_SET);
	    if ((bank->system_is_625 = (frame[3] & 0x80)) == 0) {
		bank->timeout = &timeout_60;
		bank->frame_size = (DIF_SEQ_SIZE * NUM_DIF_SEQ_525);
	    } else {
		bank->timeout = &timeout_50;
		bank->frame_size = (DIF_SEQ_SIZE * NUM_DIF_SEQ_625);
	    }
	    fstat(fileno(bank->fp), &stat);
	    bank->num_frame = stat.st_size / bank->frame_size;
	    bank->last_pos = 0L;
	    cons->show_stream_info(bank);
	    for (cnt = 0; cnt < NUM_PGM; cnt++) {
		bank->cut_pos[cnt].in = EOF;
		bank->cut_pos[cnt].out = EOF;
	    }
	}
    }

    if (bank->fp == NULL && bank->filename != NULL) {
	free(bank->filename);
	bank->filename = NULL;
    }

    return bank->fp;
}


struct key_table {
    int chr;
    int action;
};

struct key_table key_table[] = {
    { 'p', ACT_PLAY },
    { '8', ACT_PLAYx8 },
    { '4', ACT_PLAYx4 },
    { '2', ACT_PLAYx2 },
    { '1', ACT_PLAYx1 },
    { '9', ACT_PLAYs5 },
    { '0', ACT_PLAYs10 },
    { ')', ACT_BACKs10 },
    { '(', ACT_BACKs5 },
    { '!', ACT_BACKx1 },
    { '@', ACT_BACKx2 },
    { '$', ACT_BACKx4 },
    { '*', ACT_BACKx8 },
    { 's', ACT_STOP },
    { 'n', ACT_NEXT },
    { 'b', ACT_PREV },
    { 'r', ACT_REFRESH },
    { 'o', ACT_OPEN },
    { 'A', ACT_OPEN_1 },
    { 'B', ACT_OPEN_2 },
    { 'C', ACT_OPEN_3 },
    { 'D', ACT_OPEN_4 },
    { 'E', ACT_OPEN_5 },
    { 'F', ACT_OPEN_6 },
    { 'G', ACT_OPEN_7 },
    { 'H', ACT_OPEN_8 },
    { 'v', ACT_DISP },
    { 'q', ACT_QUIT },
    { 'x', ACT_QUIT },
};


int
get_key_event(FILE *fi)
{
    int c;
    struct key_table *ptr;

    c = fgetc(fi);
    for (ptr = key_table; 0 < ptr->chr; ptr++) {
	if (c == ptr->chr) {
	    return ptr->action;
	}
    }
    return ACT_NONE;
}


void
parse_subcode(u_char *frame, int system_is_625)
{
    int seq, blk, cnt, tmp;
    int num_seq = system_is_625 ? NUM_DIF_SEQ_625 : NUM_DIF_SEQ_525;
    int timecode_f = 0, datacode_f = 0, src_f = 0, date_f = 0, time_f = 0;
    u_char *DIF = frame, *ptr, *PC;
    char *source = NULL, buf[10], date[25], time[15];

    for (seq = 0; seq < num_seq; seq++) {
	for (blk = 0; blk < DIF_NUM_BLKS_IN_SEQ; blk++) {
	    switch ((DIF[0] >> 5) & 0x7) {
	      case 0:	/* Header */
		break;
	      case 1:	/* Subcode */
		if (!timecode_f) {
		    for (ptr = &DIF[3], cnt = 0; cnt < 6; cnt++, ptr += 8) {
			PC = &ptr[3];
			if (PC[0] == 0x13) {
			    /* Timecode pack */
			    cons->disp_timecode(((PC[4] >> 4) & 0x03) * 10
						+ (PC[4] & 0x0f),
						((PC[3] >> 4) & 0x07) * 10
						+ (PC[3] & 0x0f),
						((PC[2] >> 4) & 0x07) * 10
						+ (PC[2] & 0x0f),
						((PC[1] >> 4) & 0x03) * 10
						+ (PC[1] & 0x0f),
						system_is_625,
						(PC[1] & 0x40));
			    ++timecode_f;
			    break;
			}
		    }
		}
		break;
	      case 2:	/* VAUX */
		if (!datacode_f) {
		    for (ptr = &DIF[3], cnt = 0; cnt < 15; cnt++, ptr += 5) {
			PC = ptr;
			if (PC[0] == 0x60) {
			    /* VAUX Source pack */
			    if (!src_f) {
				tmp = PC[2] & 0x07;
				if (PC[1] == 0xff && tmp == 0x07) {
				    source = "---";
				} else {
				    tmp = tmp * 100
					+ (PC[1] >> 4) * 10
					+ (PC[1] & 0x0f);
				    switch (PC[3] >> 6) {
				      case 0:	/* Camera */
					source = "Camera";
					break;
				      case 1:	/* Line */
					sprintf(buf, "LINE %d", tmp);
					source = buf;
					break;
				      case 2:	/* CATV */
					sprintf(buf, "CH C%d", tmp);
					source = buf;
					break;
				      case 3:	/* TV */
					sprintf(buf, "CH %d", tmp);
					source = buf;
					break;
				    }
				}
				++src_f;
			    }
			} else if (PC[0] == 0x62) {
			    /* VAUX Rec date pack */
			    if (!date_f) {
				sprintf(date, "%x%x/%02x/%d%x",
				       (PC[4] >> 4) & 0x0f,
					PC[4] & 0x0f,
					PC[3] & 0x0f,
				       (PC[2] >> 4) & 0x03,
					PC[2] & 0x0f);
				++date_f;
			    }
			} else if (PC[0] == 0x63) {
			    /* VAUX Rec time pack */
			    if (!time_f) {
				if ((PC[1] & 0x3f) == 0x3f) {
				    tmp = 0;
				} else {
				    tmp = ((PC[1] >> 4) & 0x03) * 10
					+ (PC[1] & 0x0f);
				}
				sprintf(time, "%d%x:%d%x:%d%x.%02d",
					(PC[4] >> 4) & 0x03,
					 PC[4] & 0x0f,
					(PC[3] >> 4) & 0x07,
					 PC[3] & 0x0f,
					(PC[2] >> 4) & 0x07,
					 PC[2] & 0x0f,
					 tmp);
				++time_f;
				break;
			    }
			}
			if (src_f && date_f && time_f) {
			    cons->disp_datacode(date, time, source);
			    ++datacode_f;
			    break;
			}
		    }
		}
		break;
	      case 3:	/* Audio */
		break;
	      case 4:	/* Video */
		break;
	    }
	    DIF += DIF_BLK_SIZE;
	}
    }
}


void
usage(char *argv0)
{
    fprintf(stderr, "Usage:\t%s [options] [stream file]\n", argv0);
    fprintf(stderr, "\toptions:\n");
    fprintf(stderr, "\t   -c [<controller path>][:<controller type>]\n");
    fprintf(stderr, "\t\tUSB HID controller $B$N(B path $B$d(B type $B$r;XDj$9$k(B\n");
    fprintf(stderr, "\t\tPath $B$H(B type $B$O$I$A$i$+$,>JN,2DG=(B\n");
    fprintf(stderr, "\t\tDefault $B$O$=$l$>$l(B `%s' $B$H(B `%s'\n",
					UHID_DEVNAME, UHID_DEVTYPE);
    fprintf(stderr, "\n");
    fprintf(stderr, "\t   -n\tPlaydv $B$K$h$k2hLL=PNO$rM^;_$9$k(B(debug $BMQ!K(B\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t   -p <path>\n");
    fprintf(stderr, "\t\tPlaydv program $B$N(B path $B$r;XDj$9$k(B\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t   -v\tStream $B$N:F@8Cf$K(B timecode $B$d(B datacode $B$rI=<($9$k(B\n");
    fprintf(stderr, "\t   -w\tConsole $B$G$NI=<($K(B curses $B$r;HMQ$9$k(B\n");

    exit(1);
}
